/**
*
 
* Zilog Modbus TCP/IP Server
 
* Copyright (c), Andrzej Sokulski, 
 
* andrzej@modbus.pl - www.modbus.pl
 * All rights reserved.
 
* 
 
* 
 
*  This software is free you can redistribute it and/or
 
*  modify it under the terms of the GNU Lesser General Public
 
*  License as published by the Free Software Foundation; either
 
*  version 2.1 of the License, or (at your option) any later version.
 
*
 
*  This software is distributed in the hope that it will be useful,
 
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
 
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 
*  Lesser General Public License for more details. 
 
*  In no event shall the regents or contributors be liable for any direct,
 
*  indirect, incidential, special, exemplary, or consequential damages. 

 
*  Redistributions of source code should retain the above copyright notice,
 
*  If you want to remove the copyright notice, 
 
*  please donate me (www.modbus.pl/donate.html).
 
*
 
*/ 


#include "modbus.h"
#include <ez80.h>

void checkModbusFrame(unsigned char *ptrFrame, unsigned int size) // Check the modbus frame type
{
 if (*(ptrFrame + 6) != SlaveAddress) {
   wrongModbusFrame(ptrFrame); 
   return; 
   }else

 switch (*(ptrFrame +7))
        {
		case 1:
        readCoilModbusFrame(ptrFrame,size);
		return;
		case 2:
		readInputModbusFrame(ptrFrame,size);
		return;
		case 3:
        readRegModbusFrame(ptrFrame,size);
		return;
		case 4:
		readIRegModbusFrame(ptrFrame,size);
		return;
		case 5:
		presetCoilModbusFrame(ptrFrame,size);
		return;
		case 6:
		presetRegModbusFrame(ptrFrame,size);
		return;
		case 15:
		writeCoilModbusFrame(ptrFrame,size);
		return;
		case 16:
		writeRegModbusFrame(ptrFrame,size);
		return;
 		default:
        wrongModbusFrame(ptrFrame);
 	    return;
	  }
}

void setBits(unsigned char* srcTable, unsigned char* destTable, unsigned int destStartBit, unsigned int nrOfBits) //Sets bits in array of char
{
unsigned int pointerToChar;
unsigned int pointerToBit;
unsigned int i;
unsigned char mask;
mask = 1;
i = 0;
while (nrOfBits){
pointerToChar = destStartBit/8 ;
pointerToBit = destStartBit%8;

if (*(srcTable + i) & mask ){
*(destTable + pointerToChar)|=(1<<pointerToBit);
} else
*(destTable + pointerToChar)&=(0xFF - (1<<pointerToBit));
destStartBit ++;
nrOfBits--;
mask = mask<<1;
if (mask == 0){
   mask = 1;
   i++;
   }
}
}


void getBits(unsigned char* srcTable, unsigned char* destTable, unsigned int srcStartBit, unsigned int nrOfBits)//Gets bits from array of char
{

unsigned int pointerToChar;
unsigned int pointerToBit;
unsigned int i;
unsigned char mask;
mask = 1;
i = 0;
while (nrOfBits){
pointerToChar = srcStartBit/8 ;
pointerToBit = srcStartBit%8;

if (*(srcTable + pointerToChar) & (1<<pointerToBit) ){
*(destTable + i)|= mask;
} else
*(destTable + i)&=(0xFF - mask);
srcStartBit ++;
nrOfBits--;
mask = mask<<1;
if (mask == 0){
   mask = 1;
   i++;
   }
}
}

void wrongModbusFrame(unsigned char* ptrFrame) //Response for wrong modbus frame from client
{
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}


void presetCoilModbusFrame(unsigned char* ptrFrame,unsigned int size) // Modbus function 5
{
int maxReg,i,byteCount;
maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9);
if ((maxReg > MAX_ZILOG_COIL)|(maxReg < 0) | (12 != size)| (*(ptrFrame+11)!=0) | ((*(ptrFrame+10)!=0)& (*(ptrFrame+10)!=0xFF)) ){
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
	maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9); 
    setBits(ptrFrame+10, (unsigned char*)ZILOG_COILS, maxReg, 1);
	*(ptrFrame+5) = 6;
	LEDMATRIX_ROW = ZILOG_COILS[0];
	LEDMATRIX_COLUMN = ZILOG_COILS[1];
}
}


void writeCoilModbusFrame(unsigned char* ptrFrame,unsigned int size) // Modbus function 15
{
int maxReg,i,byteCount;
byteCount = *(ptrFrame+10)*256 + *(ptrFrame+11)/8;
if ((*(ptrFrame+10)*256 + *(ptrFrame+11))%8) byteCount++;
maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9)+*(ptrFrame+10)*256 + *(ptrFrame+11);
if ((maxReg > MAX_ZILOG_REG)|(maxReg < 1) | ((byteCount + 13) != size)){ 
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
	maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9); 
    setBits(ptrFrame+13, (unsigned char*)ZILOG_COILS, maxReg, *(ptrFrame+10)*256 + *(ptrFrame+11));
			*(ptrFrame+12) = byteCount;
	*(ptrFrame+5) = 6;
	LEDMATRIX_ROW = ZILOG_COILS[0];
	LEDMATRIX_COLUMN = ZILOG_COILS[1];

}
}

void readCoilModbusFrame(unsigned char* ptrFrame, unsigned int size) // Modbus function 1
{
unsigned int maxReg,i;
maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9)+*(ptrFrame+10)*256 + *(ptrFrame+11);
if ((maxReg > MAX_ZILOG_COIL)|(size != 12 )){ 
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
    maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9);
	i = *(ptrFrame+10)*256 + *(ptrFrame+11)/8;
    if ((*(ptrFrame+10)*256 + *(ptrFrame+11))%8 >0) i++;
   	getBits((unsigned char*)ZILOG_COILS , ptrFrame + 9,maxReg, *(ptrFrame+10)*256 + *(ptrFrame+11));
	*(ptrFrame+5) = i+3;
	*(ptrFrame+8) = i;
}
}


void readInputModbusFrame(unsigned char* ptrFrame, unsigned int size)// Modbus function 2
{
unsigned int maxReg,i;

maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9)+*(ptrFrame+10)*256 + *(ptrFrame+11);
if ((maxReg > MAX_ZILOG_INPUT)||(size != 12 )){ //Sprawdzenie poprawnoci ramki z ADAM
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
    ZILOG_INPUTS[0] = PB_DR;
    i = *(ptrFrame+10)*256 + *(ptrFrame+11)/8;
    if ((*(ptrFrame+10)*256 + *(ptrFrame+11))%8) i++;
	maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9);
	getBits((unsigned char*)ZILOG_INPUTS , ptrFrame + 9,maxReg,  *(ptrFrame+10)*256 + *(ptrFrame+11));
	*(ptrFrame+8) = i;
	*(ptrFrame+5) = i+3;

}
}



void presetRegModbusFrame(unsigned char* ptrFrame,unsigned int size)// Modbus function 6
{
int maxReg,i;
maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9);
if ((maxReg > MAX_ZILOG_REG)|(12 != size)){ 
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
	 maxReg*=2;
	 ZILOG_REG[maxReg] = *(ptrFrame+10);   
 	 ZILOG_REG[++maxReg] = *(ptrFrame+11); 
	*(ptrFrame+5) = 6;
}
}





void writeRegModbusFrame(unsigned char* ptrFrame,unsigned int size) // Modbus function 16
{
int maxReg,i;
maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9)+*(ptrFrame+10)*256 + *(ptrFrame+11);
if ((maxReg > MAX_ZILOG_REG)|(maxReg < 1) | (((maxReg-(*(ptrFrame+8)*256 + *(ptrFrame+9)))*2 + 13) != size)|(*(ptrFrame+11)>125)){ 
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
	maxReg = (*(ptrFrame+8)*256 + *(ptrFrame+9))*2; 
	for (i = 0; i<*(ptrFrame+12); i++) {
	 ZILOG_REG[maxReg] = *(ptrFrame+13+i);
 	maxReg++;
	}
	*(ptrFrame+5) = 6;
}
}



void readRegModbusFrame(unsigned char* ptrFrame,unsigned int size)// Modbus function 3
{
int maxReg,i,regCount;
maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9) +*(ptrFrame+10)*256 +*(ptrFrame+11);
if ((maxReg > MAX_ZILOG_REG)|(size!=12)|(*(ptrFrame+11)>125)){
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
	maxReg =( *(ptrFrame+8)*256 + *(ptrFrame+9))*2;
    regCount =(*(ptrFrame+10)*256 +*(ptrFrame+11))*2;
	for (i = 0; i<regCount; i++) {
	 *(ptrFrame+9+i) = ZILOG_REG[maxReg];       
 	maxReg++;
	}
	maxReg = i;
	*(ptrFrame+8) = maxReg;
	*(ptrFrame+5) = maxReg+3;
	return;
}
}

void readIRegModbusFrame(unsigned char* ptrFrame,unsigned int size)// Modbus function 4
{
int maxReg,i,regCount;
maxReg = *(ptrFrame+8)*256 + *(ptrFrame+9) +*(ptrFrame+10)*256 +*(ptrFrame+11);
if ((maxReg > MAX_ZILOG_IREG)|(size!=12)){ 
	*(ptrFrame+7) = *(ptrFrame + 7) + 0x80;
	*(ptrFrame+5) = 3;
	*(ptrFrame+8) = 2;
}else {
	maxReg = (*(ptrFrame+8)*256 + *(ptrFrame+9))*2; 
	regCount =(*(ptrFrame+10)*256 +*(ptrFrame+11))*2;
	for (i = 0; i<regCount; i++) {
	 *(ptrFrame+9+i)= ZILOG_IREG[maxReg];     
 	maxReg++;
	}
	maxReg = i+1;
	*(ptrFrame+8) = maxReg;
	*(ptrFrame+5) = maxReg+3;
	return;
}
}
