/******************************************************************************
 * File       : semaphore_test.c
 * Scope      : The scope is to demonstrate how to write an application
 *              that uses semaphore to synchronize two threads 
 *
 * Description: This sample program creates the two threads in round robin mode
 *              with a round robin time as 3 tick and also creates a binary semaphore.
 *              The thread1 acquire a semaphore while it fills a buffer with some data
 				and release the semaphore once the updating is completed.
 *              The thread2  will  acquire a semaphore ,while it reads the buffer and releases 
 				when  execution is  completed.

 *  Copyright 2004 ZiLOG inc. ALL RIGHTS RESERVED

 * The source code in this file was written by an
 * authorized ZiLOG employee or a licensed consultant.
 * The source code has been verified to the fullest
 * extent possible.
 * Permission to use this code is granted on a royalty-free
 * basis. However users are cautioned to authenticate the
 * code contained herein.
 * ZiLOG DOES NOT GUARANTEE THE VERACITY OF THE SOFTWARE.
 ************************************************************************/

//standard header files 
#include <stdio.h>				
#include <string.h>

//RZK header files 
#include "ZSysgen.h"
#include "ZTypes.h"
#include "ZThread.h"
#include "ZRegion.h"
#include"ZSemaphore.h"
#include "semaphore_test.h"
// macros for threads 
#define PRIORITY		21		/** thread priority */
#define STACK_SIZE		512   	/** stack size for the thread */
#define RR_TICK			3		/** round robin tick for the schedular */
#define MAX_SEM_COUNT	1		// create a Binary semaphore
#define MAX_LENGTH 		52		//Maximum length of the array element


#define  EN_SEMAPHORE			//Define this to use Semaphore 


//Function for initializing the UART port for printing the messages on console 
extern void InitUART();
extern RZK_REGIONHANDLE_t hGlobalRegion ;

//extern void RZKFormatError(UINT num) ;
//extern RZK_STATUS_t RZKGetErrorNum( ) ;

void Thread1();
void Thread2();
void init_array();
struct threadarguments *tArg;
//Global variables 
//thread handles to store 
RZK_THREADHANDLE_t g_hthd1 ;
RZK_THREADHANDLE_t g_hthd2 ;

RZK_SEMAPHOREHANDLE_t hSemaphore;		// Semaphore Handle 
RZK_PTR_t *threadArgs;
//stack for the two threads 
char g_thd1stack [ STACK_SIZE ];
char g_thd2stack [ STACK_SIZE ];
char msg[];
int threadEnd;

//RZK application entry 
void RZKEntry()
{

	RZK_STATUS_t status ;

	//Initialize the UART port to print the messages on console 
	InitUART();
	init_array();
   threadArgs = (RZK_PTR_t *)RZKAllocSegment( hGlobalRegion,(2 * sizeof(struct threadarguments)),INFINITE_SUSPEND );
		if( threadArgs == NULL )
		{
			printf("\nFailed to allocate segment for udpserver ");
			RZKFormatError( status ) ;
		}

	printf("\nStarting to create RZK threads" ) ; 
	tArg = (struct threadarguments *)RZKAllocSegment( hGlobalRegion,sizeof(struct threadarguments),INFINITE_SUSPEND );
		if( tArg == NULL )
		{
			printf("\nFailed to allocate segment for udpserver ");
			RZKFormatError( status ) ;
 	    }
   
	    tArg->socket = 100;
		tArg->nr = 11;
		

		threadArgs[0] = (RZK_PTR_t) tArg;
	    threadArgs[1] = NULL; 		

	/**
	 * create two threads with the same priority and the attributes of the 
	 * the thread as round robin mode
	 */
	printf("\nCreating thread #1" ) ;
	g_hthd1 = RZKCreateThread( ( unsigned char *) "Thread1",
					Thread1,
					threadArgs,
					NULL,
					( g_thd1stack + STACK_SIZE ),
					PRIORITY,
					RR_TICK, 
					//RZK_THREAD_AUTOSTART | 
					RZK_THREAD_PREEMPTION | RZK_THREAD_ROUNDROBIN ) ; 
	//validate the thread handle
	if(g_hthd1 == NULL)
	{
		printf("\nUnable to create the thread #1, error description is"); 
		RZKFormatError(RZKGetErrorNum()) ;
		return ;
	}

	printf("\nCreating thread #2" ) ;
	g_hthd2 = RZKCreateThread( ( unsigned char *) "Thread2",
					Thread2,
					NULL,
					NULL,
					( g_thd2stack + STACK_SIZE ),
					PRIORITY,
					RR_TICK, 
					//RZK_THREAD_AUTOSTART | 
					RZK_THREAD_PREEMPTION | RZK_THREAD_ROUNDROBIN ) ; 

	//validate the thread handle
	if(g_hthd2 == NULL)
	{
		printf("\nUnable to create the thread #2, error code is : %d", RZKGetErrorNum() ) ;

		//now unable to create the thread #2, delete the thread #1 
		status = RZKDeleteThread(g_hthd1) ;
		if(status != RZKERR_SUCCESS) 
		{
			printf("\nUnable to delete the thread #1, error description is");
			RZKFormatError( status ) ;
		}
		return ;
	 }
		//create a Binary semaphore
		printf(" \nCreating a Binary semaphore\n ");
		hSemaphore = RZKCreateSemaphore(( unsigned char *) " TEST ",MAX_SEM_COUNT ,RECV_ORDER_FIFO);
		if(hSemaphore == NULL)
			{
				printf("\nUnable to create the Semaphore, error description is");
				RZKFormatError(RZKGetErrorNum()) ;
			}

}
 
//Thread entry function for thread1 
void Thread1(struct threadarguments *localthread) 
{
    void **ptr;
    int i;
	
	int k=0;
	ptr = RZKGetArgPtr();
	if(ptr == NULL)
	{
		printf("\nFailed to get arguments in thread #1");
		return;
	}

	localthread = (struct threadarguments*) (*ptr);

	while(k<=10)
	{
		
		#ifdef  EN_SEMAPHORE
		RZKAcquireSemaphore(hSemaphore,MAX_INFINITE_SUSPEND);	
		#endif
		i = localthread->nr;
		printf("\Thread #1 - Nr = %d",i);
		printf("\Thread #1 - socket = %d",localthread->socket);
		printf("\nStarted Thread #1 - loop %d \n",k);
		for (i=0;i<MAX_LENGTH;i++)
			{
				
				if(k==0)
				{
					msg[i]= 'A'+i;				// Fill the array from 'A'to 'Z' 		
					if(msg[i] > 'Z')
						{
							msg[i] = 'A'+i+6;	//and then from 'a' to 'z'
						}
				 }
				 else
				 {
						msg[i]= 'z'-i;			// Fill the array from'a'to 'z'		
						if(msg[i]< 'a')
						{
							msg[i] = 'z'-i-6;
						}
				 }
				printf("T1%c:" ,msg[i]);	    //and then by 'A' to 'Z'
			}
		#ifdef  EN_SEMAPHORE
		RZKReleaseSemaphore(hSemaphore);	
		#endif
		k++;
	}
	threadEnd = 10;
}

// Thread entry function for thread2 
void Thread2()
{

	int i;
	int k=0;
	while(k<=10)
	{
	
		#ifdef  EN_SEMAPHORE
		RZKAcquireSemaphore(hSemaphore,MAX_INFINITE_SUSPEND);	
		#endif
		printf("\nStarted Thread #2 loop - %d\n",k);
		printf("\n ");
		for (i=0;i<MAX_LENGTH;i++)
			{
				printf("T2%c: " ,msg[i]);		//T2 indicates that the Thread 2 is executing

			}
		#ifdef  EN_SEMAPHORE
		RZKReleaseSemaphore(hSemaphore);	
		#endif
		k++;	
	}
}


// This routine initializes  a char array msg with 0

void init_array()
	{
		int i;
		for(i=0; i<MAX_LENGTH;i++)
			{
				msg[i] = 0;
			}
	}

//End of file
void ResumeThreads(){
RZK_STATUS_t status ;
status = RZKResumeThread(g_hthd2);
	if(status != RZKERR_SUCCESS) 
		{
			printf("\nUnable to resume the thread #2, error description is");
			RZKFormatError( status ) ;
		}
	 else
	 	printf("\nResume the thread #2 - OK");
status = RZKResumeThread(g_hthd1);
if(status != RZKERR_SUCCESS) 
		{
			printf("\nUnable to resume the thread #1, error description is");
			RZKFormatError( status ) ;
		}
	 else
	 	printf("\nResume the thread #1 - OK");

}
 

