/**
*
 
* Zilog Modbus TCP/IP Server
 
* Copyright (c), Andrzej Sokulski, 
 
* andrzej@modbus.pl - www.modbus.pl
 * All rights reserved.
 
* 
 
* 
 
*  This software is free you can redistribute it and/or
 
*  modify it under the terms of the GNU Lesser General Public
 
*  License as published by the Free Software Foundation; either
 
*  version 2.1 of the License, or (at your option) any later version.
 
*
 
*  This software is distributed in the hope that it will be useful,
 
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
 
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 
*  Lesser General Public License for more details. 
 
*  In no event shall the regents or contributors be liable for any direct,
 
*  indirect, incidential, special, exemplary, or consequential damages. 

 
*  Redistributions of source code should retain the above copyright notice,
 
*  If you want to remove the copyright notice, 
 
*  please donate me (www.modbus.pl/donate.html).
 
*
 
*/ 


//standard header files 
#include <stdio.h>				
#include <string.h>
#include <ztpconfig.h>
//RZK header files 
#include "ZSysgen.h"
#include "ZTypes.h"
#include "ZThread.h"
#include "ZRegion.h"
#include"ZSemaphore.h"
#include "basetypes.h"
#include "emulator.h"
#include "shell.h"
#include "Xc_lib.h"
#include "http.h"
#include "date.h"
#include "igmp.h"
#include "snmp.h"
#include "telnet_api.h"
#include "Telnet_Shell.h"
#include "FtpApi.h"
#include "FS_Shell.h"
#include "rtc.h"
#include "zfsapi.h"
#include "tcpstat.h"
#include "XtlTcp.h"
#include "modbus.h"
// macros for threads 
#define PRIORITY		21		/** thread priority */
#define STACK_SIZE		1028  	/** stack size for the thread */
#define RR_TICK			3		/** round robin tick for the schedular */
#define MAX_SEM_COUNT	1		// create a Binary semaphore
#define MAX_LENGTH 		52		//Maximum length of the array element


#define  EN_SEMAPHORE			//Define this to use Semaphore 


//Function for initializing the UART port for printing the messages on console 
extern void InitUART();
extern RZK_REGIONHANDLE_t hGlobalRegion ;

extern void RZKFormatError(UINT num) ;
extern RZK_STATUS_t RZKGetErrorNum( ) ;

struct tcp_nsocket{
DID TCPSrvrDevID;
DID TCPConnDevID;
char *thd1stack;
};
void Thread1();


struct threadarguments *tArg;


RZK_SEMAPHOREHANDLE_t hSemaphore;		// Semaphore Handle 
RZK_PTR_t *threadArgs;
//stack for the two threads 


void RZKTCPEntry(DID tcpServer,DID tcpConnection) // Create thread for modbus communnication
{
    RZK_THREADHANDLE_t g_hthd1 ;
    char *thd1stack;
	RZK_STATUS_t status ;
	struct tcp_nsocket *tArg;
	//Initialize the UART port to print the messages on console 
	InitUART();
	threadArgs = (RZK_PTR_t *)RZKAllocSegment( hGlobalRegion,(2 * sizeof(struct tcp_nsocket)),INFINITE_SUSPEND );
		if( threadArgs == NULL )
		{
			printf("\nFailed to allocate segment for modbus tcp server ");
			RZKFormatError( status ) ;
			return;
		}

	tArg = (struct tcp_nsocket *)RZKAllocSegment( hGlobalRegion,sizeof(struct tcp_nsocket),INFINITE_SUSPEND );
		if( tArg == NULL )
		{
			printf("\nFailed to allocate segment for modbus tcp server ");
			RZKFormatError( status ) ;
			return;
 	    }
    thd1stack = (CHAR *)RZKAllocSegment( hGlobalRegion,STACK_SIZE ,INFINITE_SUSPEND );
	if( thd1stack == NULL )
	{
			printf("\nFailed to allocate stack for modbus tcp server thread ");
			RZKFormatError( status ) ;
			return;
	}
   

	    tArg->TCPSrvrDevID = tcpServer;
        tArg->TCPConnDevID = tcpConnection;
		tArg->thd1stack = thd1stack;
		

		threadArgs[0] = (RZK_PTR_t) tArg;
	    threadArgs[1] = NULL; 		

	g_hthd1 = RZKCreateThread( ( unsigned char *) "Thread1",
					Thread1,
					threadArgs,
					NULL,
					( thd1stack + STACK_SIZE ),
					PRIORITY,
					RR_TICK, 
					RZK_THREAD_AUTOSTART | RZK_THREAD_PREEMPTION | RZK_THREAD_ROUNDROBIN ) ; 
	//validate the thread handle
	if(g_hthd1 == NULL)
	{
		printf("\nUnable to create the modbus tcp thread !"); 
		RZKFormatError(RZKGetErrorNum()) ;
		return ;
	}


	printf(" \nCreating a Binary semaphore\n "); // Semaphore for threads 
		hSemaphore = RZKCreateSemaphore(( unsigned char *) "Modbus ",MAX_SEM_COUNT ,RECV_ORDER_FIFO);
		if(hSemaphore == NULL)
			{
				printf("\nUnable to create the Semaphore, error description is");
				RZKFormatError(RZKGetErrorNum()) ;
		}
}
 
//Modbus TCP Thread 

void Thread1(struct tcp_nsocket *localthread) 
{
struct tcpstat TCPStat; 
RZK_STATUS_t retVal ;
DID TCPSrvrDevID;
DID	TCPConnDevID;
    void **ptr;
    int i;
	INT16 size = 0;
	INT16 currentLength = 0; 
	INT16 ModbusFrameLength;
    CHAR *buffer;
	CHAR *modbusFrame;
	int k=0;
	ptr = RZKGetArgPtr();
	if(ptr == NULL)
	{
		printf("\nFailed to get arguments in tcp thread #1");
		return;
	}
   buffer = (CHAR*)RZKAllocSegment( hGlobalRegion,500,INFINITE_SUSPEND );
	if( buffer == NULL )
	{
		printf("\n Error while allocating memory");
	
	while( ( retVal = RZKFreeSegment(hGlobalRegion, localthread->thd1stack ) ) != RZKERR_SUCCESS);
    while( ( retVal = RZKFreeSegment(hGlobalRegion, ptr) ) != RZKERR_SUCCESS);
	close(TCPConnDevID);
	return;
	}

 	modbusFrame = (CHAR*)RZKAllocSegment( hGlobalRegion,500,INFINITE_SUSPEND );
	if( modbusFrame == NULL )
	{
	printf("\n Error while allocating memory");
	while( ( retVal = RZKFreeSegment(hGlobalRegion, localthread->thd1stack ) ) != RZKERR_SUCCESS);
    while( ( retVal = RZKFreeSegment(hGlobalRegion, ptr) ) != RZKERR_SUCCESS);
	while(( retVal = RZKFreeSegment(hGlobalRegion, buffer) ) != RZKERR_SUCCESS);
	close(TCPConnDevID);
	return;
	}
	
	localthread = (struct tcp_nsocket*) (*ptr);
	TCPSrvrDevID = localthread->TCPSrvrDevID;
	TCPConnDevID = localthread->TCPConnDevID;

//Main loop in modbus tcp thread
while (size>=0){
	size = read(TCPConnDevID, buffer , sizeof(buffer)); // Read modbus tcp frame
	i = 0;
	if (size == SYSERR){
    	kprintf( "Error on read %X\n", size); 
    } else
       {
	   if (size > 0){
			for (i=0;i<size;i++)
			modbusFrame[currentLength + i] = buffer[i];
			currentLength +=size;
			if (currentLength > 5)
				ModbusFrameLength = (unsigned char)*(modbusFrame + 5) + 6;
			if (currentLength == ModbusFrameLength){
				RZKAcquireSemaphore(hSemaphore,MAX_INFINITE_SUSPEND); //Wait for semaphore
				checkModbusFrame((unsigned char*)modbusFrame,ModbusFrameLength);
				RZKReleaseSemaphore(hSemaphore);//Release semaphore for other threads
	 	   	    ModbusFrameLength = (unsigned char)*(modbusFrame + 5)+6;
		        size = write(TCPConnDevID, modbusFrame ,ModbusFrameLength);//Write modbus tcp frame
				currentLength = 0;
				ModbusFrameLength = 0;
				} 
	    if(size == SYSERR )
               kprintf( "Error on write %X\n", size);
        	}
  		}
	}
	kprintf("\nTCP device disconnected.......\n");
 	if (close(TCPConnDevID) == OK)
		 kprintf("TCP device %d closed....\n",TCPConnDevID);
    else
		kprintf("error on close srv  TCP device %d\n",TCPConnDevID);
 	kprintf("\nStart free segment OK.......\n");
	while( ( retVal = RZKFreeSegment(hGlobalRegion, localthread->thd1stack ) ) != RZKERR_SUCCESS);
 	while( ( retVal = RZKFreeSegment(hGlobalRegion, ptr) ) != RZKERR_SUCCESS);
 	while(( retVal = RZKFreeSegment(hGlobalRegion, buffer) ) != RZKERR_SUCCESS);
 	while(( retVal = RZKFreeSegment(hGlobalRegion, modbusFrame) ) != RZKERR_SUCCESS);
 	kprintf("\nFree segment OK.......\n");
}
 

